; LogoPixels - animated pixels of Silexica logo
; version 1.0.0 date 2017-01-25
; Copyright (C) 2017 Stefan Schuermans <stefan@blinkenarea.org>
; Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
; a BlinkenArea project - http://www.blinkenarea.org/

; ATtiny2313A, clock frequency: 8 MHz (internal oscillator)

; PA0: unused
; PA1: unused
; PA2: reset
; PB0: column 0 output (output, low)
; PB1: column 1 output (output, low)
; PB2: column 2 output (output, low)
; PB3: column 3 output (output, low)
; PB4: column 4 output (output, low)
; PB5: column 5 output (output, low)
; PB6: column 6 output (output, low)
; PB7: column 7 output (output, low)
; PD0: row 0 output (output, high)
; PD1: row 1 output (output, high)
; PD2: row 2 output (output, high)
; PD3: row 3 output (output, high)
; PD4: unused (output, high)
; PD5: unused (output, high)
; PD6: mode switch (input, pull-up enabled)



.INCLUDE        "tn2313def.inc"



; number of pixels
.equ    PIXELS                  =       32



; IO pins
.equ    COL_PORT                =       PORTB   ; column outputs
.equ    ROW_PORT                =       PORTD   ; row outputs
.equ    MODE_SW_PIN             =       PIND
.equ    MODE_SW_BIT             =       6



; general purpose registers
.def    TMP                     =       r16
.def    CNT                     =       r17
.def    DATA                    =       r18
.def    PIX                     =       r19
.def    VAL                     =       r20

; current mode
.def    MODE                    =       r22
.equ    MODE_ALL                =       0       ; play all animations
.equ    MODE_SINGLE             =       1       ; play single animation only
.equ    MODE_UNKNOWN            =       0xFF    ; unknown mode



.DSEG
.ORG    0x060



; current frame
FRAME:                  .BYTE   PIXELS



.MACRO breqx
        brne    NE
        rjmp    @0
NE:
.ENDM



; multiple nops
; parameters: N
.MACRO nop_N
.IF @0 > 0
        nop
        nop_N @0 - 1
.ENDIF
.ENDM



; output row (black/white)
; input: VAL = grayscale value (1..15)
;        X = ptr to pixel data (0..15)
; output: -
; changes: TMP, DATA, PIX
; cycles: 59
.MACRO ROW_BW
        ldi     PIX,8                   ; loop over pixels 0..7
ROW_BW_LOOP:
        ld      TMP,X+                  ; DATA.7 := if _not_ to turn on pixel
        cp      TMP,VAL
        ror     DATA                    ; shift pixel bits
        dec     PIX                     ; loop over pixels 0..7
        brne    ROW_BW_LOOP
        subi    XL,8                    ; XH not there on ATtiny2313
        com     DATA                    ; DATA.x := if to turn on pixel
        out     COL_PORT,DATA
.ENDM



; output row for grayscale (black/white) and wait
; input: VAL = grayscale value (1..15)
;        X = ptr to pixel data (0..15)
; output: -
; changes: TMP, DATA, PIX
; cycles: 71 + (grayscale - 1) * 64
.MACRO ROW_BW_WAIT
        ROW_BW                          ; 59 cycles
        mov             TMP,VAL
        dec             TMP
        rcall           WAIT_X64        ; TMP * 64 + 10 cycles
.ENDM



; turn off row (with same output timing as ROW_BW)
; input: -
; output: -
; changes: TMP
; cycles: 59
.MACRO ROW_OFF
        ldi     TMP,19
ROW_OFF_LOOP:
        dec     TMP
        brne    ROW_OFF_LOOP
        clr     TMP
        out     COL_PORT,TMP
.ENDM



.CSEG
.ORG    0x000
        rjmp    ENTRY                   ; RESET
        reti                            ; INT0
        reti                            ; INT1
        reti                            ; TIMER1_CAPT
        reti                            ; TIMER1_COMPA
        reti                            ; TIMER1_OVF
        reti                            ; TIMER0_OVF
        reti                            ; USART0_RX
        reti                            ; USART0_UDRE
        reti                            ; USART0_TX
        reti                            ; ANALOG_COMP
        reti                            ; PC_INT0
        reti                            ; TIMER1_COMPB
        reti                            ; TIMER0_COMPA
        reti                            ; TIMER0_COMPB
        reti                            ; USI_START
        reti                            ; USI_OVERFLOW
        reti                            ; EE_READY
        reti                            ; WDT
        reti                            ; PC_INT1
        reti                            ; PC_INT2



; code entry point
ENTRY:
; set system clock prescaler to 1:1
        ldi     TMP,1<<CLKPCE
        out     CLKPR,TMP
        ldi     TMP,0
        out     CLKPR,TMP
; initialize output ports
        ldi     TMP,0x00                ; PA[01] to output, low
        out     PORTA,TMP
        ldi     TMP,0x03
        out     DDRA,TMP
        ldi     TMP,0x00                ; PB[0-7] to output, low
        out     PORTB,TMP
        ldi     TMP,0xFF
        out     DDRB,TMP
        ldi     TMP,0x7F                ; PD[0-5] to output, high - PD6 to input, pull-up enabled
        out     PORTD,TMP
        ldi     TMP,0x3F
        out     DDRD,TMP
; initialize stack pointer
        ldi     TMP,RAMEND
        out     SPL,TMP
; enable watchdog (64ms)
        wdr
        ldi     TMP,1<<WDCE|1<<WDE
        out     WDTCR,TMP
        ldi     TMP,1<<WDE|1<<WDP1
        out     WDTCR,TMP
        wdr
; disable analog comparator
        ldi     TMP,1<<ACD
        out     ACSR,TMP
; jump to main program
        rjmp    MAIN



; wait n * 64 cycles
; input: TMP = duration of wait (0..14)
; output: -
; changes: TMP
; cycles: TMP * 64 + 10 (including rcall and ret)
WAIT_X64:
        lsl     TMP                     ; TMP *= 8
        lsl     TMP
        lsl     TMP
WAIT_X64_LOOP:                          ; TMP * 8 + 3 cycles
        cpi     TMP,0
        breq    WAIT_X64_DONE
        dec     TMP
        nop_N   3
        rjmp    WAIT_X64_LOOP
WAIT_X64_DONE:
        ret



; output row (grayscales)
; input: X = ptr to pixel data (0..15)
; output: -
; changes: TMP, DATA, PIX, VAL
; cycles: 7847 (including rcall and ret)
ROW_GRAY:
        ldi             VAL,1
ROW_GRAY_LOOP:
        ROW_BW_WAIT
        inc             VAL
        cpi             VAL, 16
        brne            ROW_GRAY_LOOP
        ROW_OFF
; done
        ret



; output a frame
; input: FRAME = pixel data (0..15)
; output: -
; changes: TMP, CNT, DATA, PIX, VAL, X
; cycles: 31437 (including rcall and ret)
; time: 3.9ms
OUT_FRAME:
        wdr
        ldi     XL,low(FRAME)           ; ptr to pixel data
                                        ;   XH not there on ATtiny2313
        ldi     CNT,0x01                ; bitmask loop over rows
OUT_FRAME_LOOP:
        mov     TMP,CNT                 ; select row
        com     TMP
        andi    TMP,0x0F                ; use 4 last bits (4 rows)
        ori     TMP,0x70                ; ensure other bits stay high
                                        ;   (bit 6 is pull-up for switch input)
        out     ROW_PORT,TMP
        rcall   ROW_GRAY                ; display row
        subi    XL,-8                   ; ptr to next row
                                        ;   XH not there on ATtiny2313
        lsl     CNT                     ; bitmask loop over 4 rows
        cpi     CNT,0x10
        brne    OUT_FRAME_LOOP
; done
        ret



; output a frame for some time
; input: FRAME = pixel data (0..15)
;        TMP = time to show frame (1..255, in 3.9 ms steps)
; output: -
; changes: TMP, PIX, VAL, X
; time: TMP * 3.9 ms
OUT_FRAME_TIME:
; output frame
        push    TMP
        push    CNT
        push    DATA
        rcall   OUT_FRAME               ; 3.9 ms
        pop     DATA
        pop     CNT
        pop     TMP
; loop
        dec     TMP
        brne    OUT_FRAME_TIME
; done
        ret



; read mode from switch and (store playlist entry number)
; input: MODE = old mode, CNT = playlist entry number
; output: MODE = new mode
; changes: TMP, DATA
MODE_READ:
; read new mode (into DATA)
        ldi     DATA,MODE_ALL
        sbic    MODE_SW_PIN,MODE_SW_BIT
        ldi     DATA,MODE_SINGLE
; mode was changed from all to single -> save playlist entry number
        cpi     MODE,MODE_ALL           ; old mode not all -> do nothing
        brne    MODE_READ_NOT_0_TO_1
        cpi     DATA,MODE_SINGLE        ; new mode not single -> do nothing
        brne    MODE_READ_NOT_0_TO_1
        sbic    EECR,EEPE               ; EEPROM write ongoing -> do nothing
        rjmp    MODE_READ_NOT_0_TO_1
        ldi     TMP,0<<EEPM1|0<<EEPM0   ; set EEPROM programming mode
        out     EECR,TMP
        ldi     TMP,0                   ; set EEPROM address
        out     EEARL,TMP
        mov     TMP,CNT                 ; set EEPROM data to playlist entry
        com     TMP                     ;   with NOTed number in upper nibble
        swap    TMP
        andi    TMP,0xF0
        or      TMP,CNT
        out     EEDR,TMP
        sbi     EECR,EEMPE              ; begin writing EEPROM
        sbi     EECR,EEPE
MODE_READ_NOT_0_TO_1:
; remember new mode (in MODE)
        mov     MODE,DATA
; done
        ret



; clear frame
; input: -
; output: -
; changes: X, FRAME, TMP
; time: short
CLEAR:
        ldi     XL,low(FRAME)           ; ptr to pixel data
                                        ;   XH not there on ATtiny2313
        clr     TMP
CLEAR_LOOP:
        st      X+,TMP                  ; clear pixel
        cpi     XL,low(FRAME)+PIXELS    ; bottom of loop
                                        ;   XH not there on ATtiny2313
        brne    CLEAR_LOOP
; done
        ret



; set frame to solid color
; input: DATA = value (0..15)
; output: -
; changes: X, FRAME
; time: short
SET_COLOR:
        ldi     XL,low(FRAME)           ; ptr to pixel data
                                        ;   XH not there on ATtiny2313
SET_COLOR_LOOP:
        st      X+,DATA                 ; set pixel value
        cpi     XL,low(FRAME)+PIXELS    ; bottom of loop
                                        ;   XH not there on ATtiny2313
        brne    SET_COLOR_LOOP
; done
        ret



; set pixel
; input: CNT = pixel number (0..41, nothing is done for 42..255)
;        DATA = value (0..15)
; output: -
; changes: X, FRAME, TMP
; time: short
SET_PIXEL:
        cpi     CNT,PIXELS              ; invalid pixel number -> done
        brsh    SET_PIXEL_END
        ldi     XL,low(FRAME)           ; ptr to pixel (base + offset)
        add     XL,CNT                  ;   XH not there on ATtiny2313
        st      X,DATA                  ; set pixel
SET_PIXEL_END:
; done
        ret



; draw worm
; input: CNT = head of worm (0..55)
; output: -
; changes: X, FRAME, TMP, DATA
; time: short
DRAW_WORM:
        cpi     CNT,PIXELS+14           ; invalid head pos -> done
        brsh    DRAW_WORM_END
        ldi     XL,low(FRAME)+1         ; ptr to before head
        add     XL,CNT                  ;   XH not there on ATtiny2313
        ldi     DATA,15                 ; head is full on
        cpi     CNT,PIXELS              ; head pos in frame -> go
        brlo    DRAW_WORM_LOOP
        mov     TMP,CNT                 ; TMP := invisible pixels
        subi    TMP,PIXELS-1
        sub     XL,TMP                  ; skip invisible pixels
        sub     DATA,TMP                ;   XH not there on ATtiny2313
DRAW_WORM_LOOP:
        st      -X,DATA                 ; set pixel, go back
        cpi     XL,low(FRAME)           ; 1st pixel -> done
        breq    DRAW_WORM_END           ;   XH not there on ATtiny2313
        dec     DATA                    ; next pixel darker
        brne    DRAW_WORM_LOOP          ; loop
DRAW_WORM_END:
; done
        ret



; draw backwards worm
; input: CNT = tail of worm (0..55)
; output: -
; changes: X, FRAME, TMP, DATA
; time: short
DRAW_BW_WORM:
        cpi     CNT,PIXELS+14           ; invalid tail pos -> done
        brsh    DRAW_BW_WORM_END
        ldi     XL,low(FRAME)+1         ; ptr to before tail
        add     XL,CNT                  ;   XH not there on ATtiny2313
        ldi     DATA,1                  ; tail is minimum on
        cpi     CNT,PIXELS              ; tail pos in frame -> go
        brlo    DRAW_BW_WORM_LOOP
        mov     TMP,CNT                 ; TMP := invisible pixels
        subi    TMP,PIXELS-1
        sub     XL,TMP                  ; skip invisible pixels
        add     DATA,TMP                ;   XH not there on ATtiny2313
DRAW_BW_WORM_LOOP:
        st      -X,DATA                 ; set pixel, go back
        cpi     XL,low(FRAME)           ; 1st pixel -> done
        breq    DRAW_BW_WORM_END        ;   XH not there on ATtiny2313
        inc     DATA                    ; next pixel brighter
        cpi     DATA,16                 ; loop
        brne    DRAW_BW_WORM_LOOP
DRAW_BW_WORM_END:
; done
        ret



; blink animation
; input: -
; output: -
; changes: X, FRAME, CNT, DATA, TMP
ANIM_BLINK:
; off
        ldi     DATA,0                  ; minimum color
        rcall   SET_COLOR               ; paint
        ldi     TMP,150                 ; show frame 585 ms
        rcall   OUT_FRAME_TIME
; on
        ldi     DATA,15                 ; maximum color
        rcall   SET_COLOR               ; paint
        ldi     TMP,150                 ; show frame 585 ms
        rcall   OUT_FRAME_TIME
; done
        ret



; fade up and down animation
; input: -
; output: -
; changes: X, FRAME, CNT, DATA, TMP
ANIM_FADE:
; fade up
        ldi     DATA,0                  ; start dark
ANIM_FADE_UP:
        rcall   SET_COLOR               ; paint
        ldi     TMP,15                  ; show frame 58.5 ms
        rcall   OUT_FRAME_TIME
        inc     DATA                    ; fade up
        cpi     DATA,15                 ; loop until almost full on
        brne    ANIM_FADE_UP
; fade down
ANIM_FADE_DOWN:
        rcall   SET_COLOR               ; paint
        ldi     TMP,15                  ; show frame 58.5 ms
        rcall   OUT_FRAME_TIME
        dec     DATA                    ; fade up
        cpi     DATA,255                ; loop until full off
        brne    ANIM_FADE_DOWN
; done
        ret



; flicker animation
; input: -
; output: -
; changes: X, FRAME, CNT, DATA, TMP
ANIM_FLICKER:
; even pixels
        rcall   CLEAR                   ; clear
        ldi     DATA,15                 ; even pixels to maximum
        ldi     CNT,0
ANIM_FLICKER_EVEN:
        rcall   SET_PIXEL
        subi    CNT,-2                  ; move two pixels
        cpi     CNT,PIXELS              ; loop
        brlo    ANIM_FLICKER_EVEN
        ldi     TMP,60                  ; show frame 234 ms
        rcall   OUT_FRAME_TIME
; odd pixels
        rcall   CLEAR                   ; clear
        ldi     DATA,15                 ; odd pixels to maximum
        ldi     CNT,1
ANIM_FLICKER_ODD:
        rcall   SET_PIXEL
        subi    CNT,-2                  ; move two pixels
        cpi     CNT,PIXELS              ; loop
        brlo    ANIM_FLICKER_ODD
        ldi     TMP,60                  ; show frame 234 ms
        rcall   OUT_FRAME_TIME
; done
        ret



; wobble animation
; input: -
; output: -
; changes: X, FRAME, CNT, DATA, TMP
ANIM_WOBBLE:
; even pixels up, odd pixels down
        ldi     DATA,0                  ; even pixels start dark
ANIM_WOBBLE_UP:
        ldi     CNT,0
ANIM_WOBBLE_UP_DRAW:
        rcall   SET_PIXEL
        inc     CNT                     ; next pixel
        ldi     TMP,0x0F                ; invert color
        eor     DATA,TMP
        cpi     CNT,PIXELS              ; loop
        brlo    ANIM_WOBBLE_UP_DRAW
        ldi     TMP,15                  ; show frame 58.5 ms
        rcall   OUT_FRAME_TIME
        inc     DATA                    ; next color: brighter
        cpi     DATA,16
        brlo    ANIM_WOBBLE_UP
; even pixels down, odd pixels up
        ldi     DATA,15                 ; even pixels start full
ANIM_WOBBLE_DOWN:
        ldi     CNT,0
ANIM_WOBBLE_DOWN_DRAW:
        rcall   SET_PIXEL
        inc     CNT                     ; next pixel
        ldi     TMP,0x0F                ; invert color
        eor     DATA,TMP
        cpi     CNT,PIXELS              ; loop
        brlo    ANIM_WOBBLE_DOWN_DRAW
        ldi     TMP,15                  ; show frame 58.5 ms
        rcall   OUT_FRAME_TIME
        dec     DATA                    ; next color: darker
        cpi     DATA,16
        brlo    ANIM_WOBBLE_DOWN
; done
        ret



; run animation
; input: -
; output: -
; changes: X, FRAME, CNT, DATA, TMP
ANIM_RUN:
        ldi     CNT,255                 ; start before 1st pixel
ANIM_RUN_LOOP:
        rcall   CLEAR                   ; clear
        ldi     DATA,15                 ; current pixel full on
        rcall   SET_PIXEL
        ldi     TMP,15                  ; show frame 58.5 ms
        rcall   OUT_FRAME_TIME
        inc     CNT                     ; next pixel
        cpi     CNT,PIXELS+1            ; loop until after last pixel
        brne    ANIM_RUN_LOOP
; done
        ret



; backwards run animation
; input: -
; output: -
; changes: X, FRAME, CNT, DATA, TMP
ANIM_BW_RUN:
        ldi     CNT,PIXELS              ; start after last pixel
ANIM_BW_RUN_LOOP:
        rcall   CLEAR                   ; clear
        ldi     DATA,15                 ; current pixel full on
        rcall   SET_PIXEL
        ldi     TMP,15                  ; show frame 58.5 ms
        rcall   OUT_FRAME_TIME
        dec     CNT                     ; previous pixel
        cpi     CNT,255                 ; loop until before 1st pixel
        brne    ANIM_BW_RUN_LOOP
; done
        ret



; worm animation
; input: -
; output: -
; changes: X, FRAME, CNT, DATA, TMP
ANIM_WORM:
        ldi     CNT,255                 ; worm starts before 1st pixel
ANIM_WORM_LOOP:
        rcall   CLEAR                   ; draw worm
        rcall   DRAW_WORM
        ldi     TMP,15                  ; show frame 58.5 ms
        rcall   OUT_FRAME_TIME
        inc     CNT                     ; advance worm
        cpi     CNT,PIXELS+15           ; loop until has exits
        brne    ANIM_WORM_LOOP
; done
        ret



; backwards worm animation
; input: -
; output: -
; changes: X, FRAME, CNT, DATA, TMP
ANIM_BW_WORM:
        ldi     CNT,PIXELS+14           ; worm starts behind frame
                                        ;   head not yet visible
ANIM_BW_WORM_LOOP:
        rcall   CLEAR                   ; draw backwards worm
        rcall   DRAW_BW_WORM
        ldi     TMP,15                  ; show frame 58.5 ms
        rcall   OUT_FRAME_TIME
        dec     CNT                     ; advance worm backwards
        cpi     CNT,254                 ; loop until worm has exited
        brne    ANIM_BW_WORM_LOOP
; done
        ret



; blink playlist entry
; input: -
; output: -
; changes: X, FRAME, CNT, DATA, TMP
PLAY_BLINK:
        rcall   ANIM_BLINK
        rcall   ANIM_BLINK
        rcall   ANIM_BLINK
        ret



; fade up and down playlist entry
; input: -
; output: -
; changes: X, FRAME, CNT, DATA, TMP
PLAY_FADE:
        rcall   ANIM_FADE
        rcall   ANIM_FADE
        ret



; flicker playlist entry
; input: -
; output: -
; changes: X, FRAME, CNT, DATA, TMP
PLAY_FLICKER:
        ldi     CNT,10
PLAY_FLICKER_LOOP:
        push    CNT
        rcall   ANIM_FLICKER
        pop     CNT
        dec     CNT
        brne    PLAY_FLICKER_LOOP
        ret



; wobble playlist entry
; input: -
; output: -
; changes: X, FRAME, CNT, DATA, TMP
PLAY_WOBBLE:
        rcall   ANIM_WOBBLE
        rcall   ANIM_WOBBLE
        rcall   ANIM_WOBBLE
        rcall   ANIM_WOBBLE
        rcall   ANIM_WOBBLE
        ret



; run playlist entry
; input: -
; output: -
; changes: X, FRAME, CNT, DATA, TMP
PLAY_RUN:
        rcall   ANIM_RUN
        rcall   ANIM_RUN
        rcall   ANIM_RUN
        ret



; backwards run playlist entry
; input: -
; output: -
; changes: X, FRAME, CNT, DATA, TMP
PLAY_BW_RUN:
        rcall   ANIM_BW_RUN
        rcall   ANIM_BW_RUN
        rcall   ANIM_BW_RUN
        ret



; worm playlist entry
; input: -
; output: -
; changes: X, FRAME, CNT, DATA, TMP
PLAY_WORM:
        rcall   ANIM_WORM
        rcall   ANIM_WORM
        rcall   ANIM_WORM
        ret



; backwards worm playlist entry
; input: -
; output: -
; changes: X, FRAME, CNT, DATA, TMP
PLAY_BW_WORM:
        rcall   ANIM_BW_WORM
        rcall   ANIM_BW_WORM
        rcall   ANIM_BW_WORM
        ret



; user animation data
USER_ANIM_DATA:
.INCLUDE        "anim.asm"
USER_ANIM_DATA_END:



; user animation playlist entry
; input: -
; output: -
; changes: X, Z, FRAME, CNT, DATA, TMP
PLAY_USER_ANIM:
; pointer to start of data
        ldi     ZL,low(USER_ANIM_DATA*2)
        ldi     ZH,high(USER_ANIM_DATA*2)
; if pointer at end of data -> leave loop
PLAY_USER_ANIM_LOOP:
        cpi     ZH,high(USER_ANIM_DATA_END*2)
        brlo    PLAY_USER_ANIM_GO
        brne    PLAY_USER_ANIM_END
        cpi     ZL,low(USER_ANIM_DATA_END*2)
        brsh    PLAY_USER_ANIM_END
PLAY_USER_ANIM_GO:
; load frame
        lpm     TMP,Z+                  ; unused byte
        lpm     TMP,Z+                  ; frame duration in 3.9 ms
        ldi     XL,low(FRAME)           ; ptr to frame buffer
                                        ;   XH not there on ATtiny2313
        ldi     CNT,PIXELS/2            ; two pixel per byte in flash data
PLAY_USER_ANIM_LOAD:
        lpm     DATA,Z                  ; load from flash, do no advance
        swap    DATA                    ; get upper nibble
        andi    DATA,0x0F
        st      X+,DATA                 ; store to frame, advance
        lpm     DATA,Z+                 ; load from flash again, advance
        andi    DATA,0x0F               ; get lower nibble
        st      X+,DATA                 ; store to frame, advance
        dec     CNT
        brne    PLAY_USER_ANIM_LOAD
; show frame
        rcall   OUT_FRAME_TIME
; next frame

        rjmp    PLAY_USER_ANIM_LOOP
; done
PLAY_USER_ANIM_END:
        ret



; playlist
PLAYLIST:
        rjmp    PLAY_USER_ANIM
        rjmp    PLAY_BLINK
        rjmp    PLAY_WORM
        rjmp    PLAY_FLICKER
        rjmp    PLAY_RUN
        rjmp    PLAY_FADE
        rjmp    PLAY_BW_WORM
        rjmp    PLAY_WOBBLE
        rjmp    PLAY_BW_RUN
PLAYLIST_END:

.equ    PLAYLIST_ENTRIES        =       PLAYLIST_END - PLAYLIST



; play playlist entry
; input: CNT = playlist entry
; output: -
; changes: X, Z, FRAME, DATA, TMP
PLAYLIST_PLAY:
; check playlist entry
        cpi     CNT,PLAYLIST_ENTRIES
        brsh    PLAYLIST_PLAY_END
; call playlist entry (save CNT register)
        clr     TMP
        ldi     ZL,low(PLAYLIST)
        ldi     ZH,high(PLAYLIST)
        add     ZL,CNT
        adc     ZH,TMP
        push    CNT
        icall
        pop     CNT
; done
PLAYLIST_PLAY_END:
        ret



; main program
MAIN:
        wdr

; initialization
        ldi     MODE,MODE_UNKNOWN       ; unknown mode

; get number of fist playlist entry from EEPROM (into CNT)
        ldi     TMP,0                   ; set EEPROM address
        out     EEARL,TMP
        sbi     EECR,EERE               ; start EEPROM read
        in      CNT,EEDR                ; get read value
        mov     TMP,CNT                 ; check if high nibble contains NOTed
        com     TMP                     ;   value
        swap    TMP
        cp      CNT,TMP
        brne    MAIN_FIRST_INVALID
        andi    CNT,0x0F                ; throw away check value in high nibble
        rjmp    MAIN_FIRST_END
MAIN_FIRST_INVALID:
        clr     CNT                     ; start with first playlist entry
MAIN_FIRST_END:

; main loop
MAIN_LOOP:
        wdr

        ; play current playlist entry (CNT)
        rcall   PLAYLIST_PLAY

        ; read new mode (playlist entry in CNT)
        rcall   MODE_READ

        ; go to next playlist entry if not in mode single
        cpi     MODE,MODE_SINGLE
        breq    MAIN_NO_NEXT
        inc     CNT
MAIN_NO_NEXT:
        ; go back to begin of playlist if beyond end
        cpi     CNT,PLAYLIST_ENTRIES
        brlo    MAIN_NO_WRAP
        clr     CNT
MAIN_NO_WRAP:

        ; bottom of main loop
        rjmp     MAIN_LOOP

